from typing import Any, Dict

from usepolvo.arms.base_resource import BaseResource
from usepolvo.beak.exceptions import ResourceNotFoundError, ValidationError
from usepolvo.tentacles.linear.resources.issues.schemas import Issue, IssueListResponse


class LinearIssueResource(BaseResource):
    def __init__(self, client):
        super().__init__(client)
        self.resource_type = "issue"

    def list(self, page: int = 1, size: int = 10) -> IssueListResponse:
        """
        List issues with GraphQL pagination.
        """
        try:
            response = self.client._request(
                "GET", self.resource_type, params={"first": size, "after": None}  # GraphQL pagination
            )

            result = IssueListResponse(**response[f"{self.resource_type}s"])
            return result
        except Exception as e:
            self.client.handle_error(e)

    def get(self, resource_id: str) -> Issue:
        """
        Get a single issue by ID.
        """
        try:
            response = self.client._request("GET", f"{self.resource_type}/{resource_id}")
            if not response.get(self.resource_type):
                raise ResourceNotFoundError(f"Issue with ID {resource_id} not found")
            return Issue(**response[self.resource_type])
        except Exception as e:
            self.client.handle_error(e)

    def create(self, data: Dict[str, Any]) -> Issue:
        """
        Create a new issue.
        """
        try:
            response = self.client._request(
                "POST", self.resource_type, json={"input": self._prepare_request_data(data)}
            )
            return Issue(**response[f"create{self.resource_type.title()}"])
        except Exception as e:
            raise ValidationError(f"Invalid data for creating issue: {str(e)}")

    def update(self, resource_id: str, data: Dict[str, Any]) -> Issue:
        """
        Update an existing issue.
        """
        try:
            response = self.client._request(
                "PUT", f"{self.resource_type}/{resource_id}", json={"input": self._prepare_request_data(data)}
            )
            return Issue(**response[f"{self.resource_type}Update"][self.resource_type])
        except Exception as e:
            self.client.handle_error(e)

    def delete(self, resource_id: str) -> bool:
        """
        Delete an issue.
        """
        try:
            response = self.client._request("DELETE", f"{self.resource_type}/{resource_id}")
            return response[f"delete{self.resource_type.title()}"]["success"]
        except Exception as e:
            self.client.handle_error(e)

    def _prepare_request_data(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Prepare data for GraphQL mutations.
        """
        # Convert snake_case to camelCase and handle any specific Linear API requirements
        prepared_data = super()._prepare_request_data(data)

        # Handle any specific Linear API field transformations
        if "team_id" in prepared_data:
            prepared_data["teamId"] = prepared_data.pop("team_id")

        return prepared_data



from functools import wraps
from typing import Optional

import stripe

from usepolvo.arms.base_client import BaseClient
from usepolvo.tentacles.stripe.config import get_settings
from usepolvo.tentacles.stripe.customers.resource import StripeCustomerResource
from usepolvo.tentacles.stripe.rate_limiter import StripeRateLimiter


class StripeClient(BaseClient):
    def __init__(self, api_key: Optional[str] = None):
        super().__init__()
        settings = get_settings()
        stripe.api_key = api_key if api_key else settings.stripe_api_key
        self.stripe = stripe
        self.rate_limiter = StripeRateLimiter()
        self._customers = None

    @property
    def customers(self):
        if self._customers is None:
            self._customers = StripeCustomerResource(self)
        return self._customers

    def rate_limited_execute(self, method, *args, **kwargs):
        @wraps(method)
        def wrapper(*args, **kwargs):
            is_write_operation = kwargs.get("is_write_operation", False)
            self.rate_limiter.wait_if_needed(is_write_operation=is_write_operation)
            return method(*args, **kwargs)

        return wrapper(*args, **kwargs)

    def handle_error(self, e):
        super().handle_error(e)
        # Additional Stripe-specific error handling can be added here



from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Tuple

from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport

from usepolvo.arms.base_client import BaseClient


class BaseGraphQLClient(BaseClient, ABC):
    """Base client for GraphQL-based API integrations."""

    def __init__(self):
        super().__init__()
        self._client = None
        self._setup_graphql_client()

    @abstractmethod
    def get_auth_headers(self) -> Dict[str, str]:
        """
        Get authentication headers for the GraphQL client.
        Must be implemented by child classes.

        :return: Dictionary of headers
        """
        pass

    def _setup_graphql_client(self):
        """Initialize the GraphQL client with authentication headers."""
        if not self.base_url:
            raise ValueError("base_url must be set before initializing the GraphQL client")

        transport = RequestsHTTPTransport(
            url=self.base_url,
            headers=self.get_auth_headers(),
            use_json=True,
        )
        self._client = Client(transport=transport, fetch_schema_from_transport=True)

    def execute_query(self, query: str, variables: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        """
        Execute a GraphQL query with error handling.

        :param query: The GraphQL query string
        :param variables: Optional variables for the query
        :return: The query result
        """
        if not self._client:
            raise RuntimeError("GraphQL client not initialized")

        try:
            result = self._client.execute(gql(query), variable_values=variables)
            return result
        except Exception as e:
            self.handle_error(e)

    def _convert_rest_to_graphql(self, method: str, endpoint: str, **kwargs) -> Tuple[str, Dict[str, Any]]:
        """
        Convert REST-style requests to GraphQL queries.

        :param method: HTTP method
        :param endpoint: The endpoint path
        :param kwargs: Additional arguments
        :return: Tuple of (query_string, variables)
        """
        parts = endpoint.strip("/").split("/")
        resource_type = parts[0]
        resource_id = parts[1] if len(parts) > 1 else None
        request_payload = kwargs.get("json", {})
        request_params = kwargs.get("params", {})

        if method == "GET":
            if resource_id:
                return self._build_get_query(resource_type), {"id": resource_id}
            return self._build_list_query(resource_type, request_params), request_params
        elif method == "POST":
            return self._build_create_mutation(resource_type), request_payload
        elif method == "PUT":
            return self._build_update_mutation(resource_type), {"id": resource_id, **request_payload}
        elif method == "DELETE":
            return self._build_delete_mutation(resource_type, resource_id), {"id": resource_id}

        raise ValueError(f"Unsupported method: {method}")

    def _build_get_query(self, resource_type: str) -> str:
        """Build a GraphQL query for fetching a single resource."""
        return f"""
            query Get{resource_type.title()}($id: String!) {{
                {resource_type}(id: $id) {{
                    id
                    {self._get_resource_fields(resource_type)}
                }}
            }}
        """

    def _build_list_query(self, resource_type: str, params: Dict[str, Any]) -> str:
        """Build a GraphQL query for listing resources."""
        return f"""
            query List{resource_type.title()}s($first: Int, $after: String) {{
                {resource_type}s(first: $first, after: $after) {{
                    nodes {{
                        id
                        {self._get_resource_fields(resource_type)}
                    }}
                    pageInfo {{
                        hasNextPage
                        endCursor
                    }}
                }}
            }}
        """

    def _build_create_mutation(self, resource_type: str) -> str:
        """Build a GraphQL mutation for creating a resource."""
        return f"""
            mutation Create{resource_type.title()}($input: Create{resource_type.title()}Input!) {{
                create{resource_type.title()}(input: $input) {{
                    id
                    {self._get_resource_fields(resource_type)}
                }}
            }}
        """

    def _build_update_mutation(self, resource_type: str) -> str:
        """Build a GraphQL mutation for updating a resource."""
        resource_title = resource_type.title()
        return f"""
            mutation {resource_title}Update($id: String!, $input: {resource_title}UpdateInput!) {{
                {resource_type}Update(id: $id, input: $input) {{
                    issue {{ id }}
                }}
            }}
        """

    def _build_delete_mutation(self, resource_type: str, resource_id: str) -> str:
        """Build a GraphQL mutation for deleting a resource."""
        return f"""
            mutation Delete{resource_type.title()}($id: ID!) {{
                delete{resource_type.title()}(id: $id) {{
                    success
                }}
            }}
        """

    def _get_resource_fields(self, resource_type: str) -> str:
        """
        Get the fields to be queried for a specific resource type.
        Should be overridden by child classes to specify fields for each resource.
        """
        return """
            title
            description
        """



from hubspot.crm.contacts import ApiException


class HubSpotError(Exception):
    """Base class for all HubSpot-related exceptions."""

    pass


class HubSpotAuthenticationError(HubSpotError):
    """Raised when there is an authentication error with HubSpot."""

    pass


class HubSpotAPIError(HubSpotError):
    """Raised when there is an API error with HubSpot."""

    pass


class HubSpotRateLimitError(HubSpotError):
    """Raised when HubSpot's rate limit is exceeded."""

    pass


def handle_hubspot_error(e: Exception):
    """Handles HubSpot exceptions and raises appropriate usepolvo exceptions."""
    if isinstance(e, ApiException):
        if e.status == 401:
            raise HubSpotAuthenticationError(f"Authentication error: {e}")
        elif e.status == 429:
            raise HubSpotRateLimitError(f"Rate limit exceeded: {e}")
        else:
            raise HubSpotAPIError(f"API error: {e}")
    else:
        raise HubSpotError(f"An unexpected error occurred: {e}")



from typing import Any, Dict, List

from pydantic import BaseModel


class IssueState(BaseModel):
    id: str
    name: str


class IssueAssignee(BaseModel):
    id: str
    name: str


class Issue(BaseModel):
    id: str
    title: str | None = None
    description: str | None = None
    state: IssueState | None = None
    assignee: IssueAssignee | None = None


class IssueListResponse(BaseModel):
    nodes: List[Issue]
    pageInfo: Dict[str, Any]



from typing import Any, Dict, Optional

from usepolvo.arms.base_webhook import BaseWebhook
from usepolvo.tentacles.linear.config import get_settings
from usepolvo.tentacles.linear.webhooks.schemas import LinearWebhookPayload


class LinearWebhook(BaseWebhook):
    def __init__(self, webhook_secret: Optional[str] = None):
        super().__init__()
        self.signature_header = "Linear-Signature"

        # Initialize settings
        self.settings = get_settings()

        # Set webhook secret from argument or fall back to settings
        secret = webhook_secret if webhook_secret else self.settings.linear_webhook_secret
        if secret:
            self.set_secret_key(secret)

    def get_event_type(self, payload: Dict[str, Any]) -> str:
        return LinearWebhookPayload(**payload).get_event_type()

    async def process(self, payload: Dict[str, Any]) -> Any:
        try:
            validated_payload = LinearWebhookPayload(**payload)
        except Exception as e:
            print(f"Invalid payload: {e}")
            raise
        event_type = validated_payload.get_event_type()
        handler = self.handlers.get(event_type, self.default_handler)
        return await handler(validated_payload)

    async def default_handler(self, payload: LinearWebhookPayload):
        """Default handler for unhandled webhook events"""
        event_type = payload.get_event_type()
        print(f"Unhandled event: {event_type}")
        return {"status": "ok"}



from typing import Optional

from anthropic import Anthropic

from usepolvo.arms.base_client import BaseClient
from usepolvo.tentacles.claude.completions.resource import ClaudeCompletionResource
from usepolvo.tentacles.claude.config import get_settings
from usepolvo.tentacles.claude.rate_limiter import ClaudeRateLimiter


class ClaudeClient(BaseClient):
    def __init__(self, api_key: Optional[str] = None):
        super().__init__()
        settings = get_settings()
        self.api_key = api_key if api_key else settings.claude_api_key
        self.client = Anthropic(api_key=self.api_key)
        self.rate_limiter = ClaudeRateLimiter()
        self._completions = None

    @property
    def completions(self):
        if self._completions is None:
            self._completions = ClaudeCompletionResource(self)
        return self._completions

    def handle_error(self, e):
        super().handle_error(e)
        # Additional Claude-specific error handling can be added here



from typing import Any, Dict, List

from stripe.error import StripeError

from usepolvo.arms.base_resource import BaseResource
from usepolvo.beak.exceptions import ResourceNotFoundError, ValidationError


class StripeCustomerResource(BaseResource):
    def __init__(self, client):
        super().__init__(client)
        self.stripe = client.stripe

    def list(
        self, page: int = 1, size: int = 10, starting_after: str = None, ending_before: str = None, **kwargs
    ) -> List[Dict[str, Any]]:
        """
        Retrieve a list of Stripe customers.

        :param page: The page number to retrieve
        :param size: A limit on the number of objects to be returned. Limit can range between 1 and 100.
        :param starting_after: A cursor for use in pagination. Defines your place in the list.
        :param ending_before: A cursor for use in pagination. Defines your place in the list.
        :param kwargs: Additional parameters for the list operation
        :return: A list of customer objects
        """
        try:
            self.client.rate_limiter.wait_if_needed(is_write_operation=False)
            params = self.client.get_pagination_params(page, size, starting_after, ending_before)
            params.update(kwargs)
            return self.stripe.Customer.list(**params)
        except StripeError as e:
            self.client.handle_error(e)

    def get(self, resource_id: str) -> Dict[str, Any]:
        """
        Retrieve a single Stripe customer by ID.

        :param resource_id: The ID of the customer to retrieve
        :return: The customer object
        :raises ResourceNotFoundError: If the customer is not found
        """
        try:
            self.client.rate_limiter.wait_if_needed(is_write_operation=False)
            return self.stripe.Customer.retrieve(resource_id)
        except self.stripe.error.InvalidRequestError:
            raise ResourceNotFoundError(f"Customer with ID {resource_id} not found")
        except StripeError as e:
            self.client.handle_error(e)

    def create(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Create a new Stripe customer.

        :param data: The data for the new customer
        :return: The created customer object
        :raises ValidationError: If the data is invalid
        """
        try:
            self.client.rate_limiter.wait_if_needed(is_write_operation=True)
            return self.stripe.Customer.create(**data)
        except self.stripe.error.InvalidRequestError as e:
            raise ValidationError(f"Invalid data for creating customer: {str(e)}")
        except StripeError as e:
            self.client.handle_error(e)

    def update(self, resource_id: str, data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Update an existing Stripe customer.

        :param resource_id: The ID of the customer to update
        :param data: The updated data for the customer
        :return: The updated customer object
        :raises ResourceNotFoundError: If the customer is not found
        :raises ValidationError: If the data is invalid
        """
        try:
            self.client.rate_limiter.wait_if_needed(is_write_operation=True)
            customer = self.stripe.Customer.modify(resource_id, **data)
            return customer
        except self.stripe.error.InvalidRequestError as e:
            if "No such customer" in str(e):
                raise ResourceNotFoundError(f"Customer with ID {resource_id} not found")
            else:
                raise ValidationError(f"Invalid data for updating customer: {str(e)}")
        except StripeError as e:
            self.client.handle_error(e)

    def delete(self, resource_id: str) -> None:
        """
        Delete a Stripe customer.

        :param resource_id: The ID of the customer to delete
        :raises ResourceNotFoundError: If the customer is not found
        """
        try:
            self.client.rate_limiter.wait_if_needed(is_write_operation=True)
            self.stripe.Customer.delete(resource_id)
        except self.stripe.error.InvalidRequestError:
            raise ResourceNotFoundError(f"Customer with ID {resource_id} not found")
        except StripeError as e:
            self.client.handle_error(e)



from typing import Any, Dict

from anthropic import Anthropic

from usepolvo.arms.base_resource import BaseResource
from usepolvo.beak.exceptions import ResourceNotFoundError, ValidationError
from usepolvo.tentacles.claude.exceptions import ClaudeError


class ClaudeCompletionResource(BaseResource):
    def __init__(self, client):
        super().__init__(client)
        self.anthropic = client.client

    def create(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Create a new completion using Claude.

        :param data: The data for the new completion
        :return: The created completion object
        :raises ValidationError: If the data is invalid
        """
        try:
            self.client.rate_limiter.wait_if_needed()
            response = self.anthropic.completions.create(**data)
            return response.model_dump()
        except ValueError as e:
            raise ValidationError(f"Invalid data for creating completion: {str(e)}")
        except Exception as e:
            self.client.handle_error(e)

    def get(self, completion_id: str) -> Dict[str, Any]:
        pass

    def list(self, **kwargs) -> Dict[str, Any]:
        pass

    def update(self, completion_id: str, data: Dict[str, Any]) -> Dict[str, Any]:
        pass

    def delete(self, completion_id: str) -> Dict[str, Any]:
        pass



# usepolvo/tentacles/salesforce/opportunities/resource.py

from typing import Any, Dict

from usepolvo.arms.base_resource import BaseResource
from usepolvo.beak.exceptions import ResourceNotFoundError, ValidationError
from usepolvo.tentacles.salesforce.opportunities.schemas import (
    OpportunityListResponse,
    OpportunityResponse,
)


class SalesforceOpportunityResource(BaseResource):
    def __init__(self, client):
        super().__init__(client)
        self.base_path = "/sobjects/Opportunity"

    def list(self, page: int = 1, size: int = 10) -> OpportunityListResponse:
        cache_key = f"list_opportunities_page_{page}_size_{size}"
        if cache_key in self.client.cache:
            return self.client.cache[cache_key]

        # Ensure the correct endpoint is being used
        try:
            response = self.client._request("GET", self.base_path)
            result = OpportunityListResponse(**response)
            self.client.cache[cache_key] = result
            return result
        except Exception as e:
            self.client.handle_error(e)

    def get(self, resource_id: str) -> OpportunityResponse:
        try:
            response = self.client._request("GET", f"{self.base_path}/{resource_id}")
            return OpportunityResponse(**response)
        except ResourceNotFoundError:
            raise ResourceNotFoundError(f"Opportunity with ID {resource_id} not found")

    def create(self, data: Dict[str, Any]) -> OpportunityResponse:
        try:
            prepared_data = self._prepare_request_data(data)
            response = self.client._request("POST", self.base_path, json=prepared_data)
            return OpportunityResponse(**response)
        except ValidationError as e:
            raise ValidationError(f"Invalid data for creating opportunity: {str(e)}")

    def update(self, resource_id: str, data: Dict[str, Any]) -> OpportunityResponse:
        try:
            prepared_data = self._prepare_request_data(data)
            response = self.client._request("PATCH", f"{self.base_path}/{resource_id}", json=prepared_data)
            return OpportunityResponse(**response)
        except ResourceNotFoundError:
            raise ResourceNotFoundError(f"Opportunity with ID {resource_id} not found")
        except ValidationError as e:
            raise ValidationError(f"Invalid data for updating opportunity: {str(e)}")

    def delete(self, resource_id: str) -> None:
        try:
            self.client._request("DELETE", f"{self.base_path}{resource_id}")
        except ResourceNotFoundError:
            raise ResourceNotFoundError(f"Opportunity with ID {resource_id} not found")



# usepolvo/tentacles/salesforce/config.py

from functools import lru_cache
from typing import Optional
from urllib.parse import urlparse

from pydantic_settings import BaseSettings


class SalesforceSettings(BaseSettings):
    salesforce_consumer_key: Optional[str] = None
    salesforce_consumer_secret: Optional[str] = None
    salesforce_redirect_uri: Optional[str] = "http://localhost:8000/callback"
    salesforce_environment: str = "production"  # Can be 'production' or 'sandbox'
    salesforce_custom_domain: Optional[str] = None
    salesforce_api_version: str = "v61.0"  # TODO: make this configurable

    @property
    def salesforce_instance_url(self):
        if self.salesforce_custom_domain:
            return f"https://{self.salesforce_custom_domain}.my.salesforce.com"
        else:
            domain = "test.salesforce.com" if self.salesforce_environment == "sandbox" else "login.salesforce.com"
            return f"https://{domain}"

    @property
    def salesforce_oauth2_url(self):
        return f"{self.salesforce_instance_url}/services/oauth2"

    @property
    def salesforce_auth_url(self):
        return f"{self.salesforce_oauth2_url}/authorize"

    @property
    def salesforce_token_url(self):
        return f"{self.salesforce_oauth2_url}/token"

    @property
    def salesforce_api_base_url(self):
        return f"{self.salesforce_instance_url}/services/data/{self.salesforce_api_version}"

    class Config:
        env_file = ".env"
        extra = "ignore"


@lru_cache
def get_settings():
    return SalesforceSettings()



# usepolvo/tentacles/certn/applications/schemas.py

from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, HttpUrl


class Applicant(BaseModel):
    id: str
    status: str
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    email: Optional[str] = None
    phone_number: Optional[str] = None
    application_url: Optional[HttpUrl] = None
    report_url: HttpUrl
    employer_references_min: int


class Application(BaseModel):
    created: datetime
    modified: datetime
    id: str
    applicant: Applicant
    is_active: bool
    is_selected: bool
    team_id: str


class EnhancedIdentityVerification(BaseModel):
    dispute: bool
    status: str
    status_label: str


class ReportSummary(BaseModel):
    enhanced_identity_verification: EnhancedIdentityVerification
    report_result: str
    report_result_label: str
    dispute: bool
    report_status: str
    report_status_label: str


class ApplicationResponse(BaseModel):
    created: datetime
    modified: datetime
    last_updated: datetime
    submitted_time: Optional[datetime]
    id: str
    short_uid: str
    is_submitted: bool
    applicant_type: str
    result: str
    result_label: str
    report_status: str
    country: str
    order_status: str
    information: Dict[str, Any]
    application: Application
    report_summary: ReportSummary
    status: str
    status_label: str


class ApplicationListResponse(BaseModel):
    count: int
    next: Optional[HttpUrl] = None
    previous: Optional[HttpUrl] = None
    results: List[ApplicationResponse]



# usepolvo/tentacles/linear/exceptions.py

import requests


class LinearError(Exception):
    """Base class for all Linear-related exceptions."""

    pass


class LinearAuthenticationError(LinearError):
    """Raised when there is an authentication error with Linear."""

    pass


class LinearAPIError(LinearError):
    """Raised when there is an API error with Linear."""

    pass


def handle_linear_error(error: requests.RequestException):
    if isinstance(error, requests.exceptions.HTTPError):
        if error.response.status_code == 401:
            return LinearAuthenticationError("Authentication failed. Please check your API key.")
        elif error.response.status_code == 403:
            return LinearAuthenticationError("Permission denied. Please check your API key permissions.")
        else:
            return LinearAPIError(f"API error: {error.response.status_code} - {error.response.text}")
    elif isinstance(error, requests.exceptions.ConnectionError):
        return LinearAPIError("Connection error. Please check your internet connection.")
    elif isinstance(error, requests.exceptions.Timeout):
        return LinearAPIError("Request timed out. Please try again later.")
    else:
        return LinearError(f"An unexpected error occurred: {str(error)}")



# usepolvo/tentacles/linear/config.py

from functools import lru_cache
from typing import Optional

from pydantic_settings import BaseSettings


@lru_cache
def get_settings():
    return LinearSettings()


class LinearSettings(BaseSettings, env_file=".env", extra="ignore"):
    linear_api_key: Optional[str] = None
    linear_client_id: Optional[str] = None
    linear_client_secret: Optional[str] = None
    linear_base_url: str = "https://api.linear.app/graphql"
    linear_webhook_secret: Optional[str] = None



# usepolvo/tentacles/certn/config.py

from functools import lru_cache
from typing import Optional

from pydantic_settings import BaseSettings


@lru_cache
def get_settings():
    return CertnSettings()


class CertnSettings(BaseSettings, env_file=".env", extra="ignore"):
    certn_api_key: Optional[str] = None
    certn_base_url: str = "https://demo-api.certn.co"  # Default API URL



from functools import lru_cache
from typing import Optional

from pydantic_settings import BaseSettings


@lru_cache
def get_settings():
    return Settings()


class Settings(BaseSettings):
    hubspot_api_key: Optional[str] = None
    hubspot_client_id: Optional[str] = None
    hubspot_client_secret: Optional[str] = None
    hubspot_redirect_uri: Optional[str] = None

    class Config:
        env_file = ".env"
        extra = "ignore"



from typing import Optional

from pydantic import BaseModel, Field


class CompletionRequest(BaseModel):
    model: str = Field(default="gemini-pro")
    prompt: str
    max_output_tokens: Optional[int] = Field(default=None, ge=1, le=2048)
    temperature: Optional[float] = Field(default=None, ge=0, le=1)
    top_p: Optional[float] = Field(default=None, ge=0, le=1)
    top_k: Optional[int] = Field(default=None, ge=1)


class CompletionResponse(BaseModel):
    text: str



import hashlib
import hmac


def verify_hmac_signature(payload: str, signature: str, secret_key: str):
    """
    Verify webhook signature using HMAC-SHA256.

    Args:
        payload: Raw request body as string
        signature: The signature from webhook header
        secret_key: The webhook secret key

    Raises:
        ValueError: If signature verification fails
    """
    # Strip quotes if present
    signature = signature.strip('"')
    secret_key = secret_key.strip('"')

    # Compute HMAC
    computed_signature = hmac.new(
        key=secret_key.encode("utf-8"), msg=payload.encode("utf-8"), digestmod=hashlib.sha256
    ).hexdigest()

    # Use constant-time comparison
    if not hmac.compare_digest(computed_signature, signature):
        raise ValueError("Invalid webhook signature")



from pydantic import BaseModel, Field, field_validator


class CompletionRequest(BaseModel):
    prompt: str
    model: str = Field(default="claude-2")
    max_tokens_to_sample: int = Field(default=1000, ge=1, le=100000)
    temperature: float = Field(default=0.7, ge=0, le=1)

    @field_validator("prompt")
    def append_to_prompt(cls, v):
        if not v.startswith("\n\nHuman:"):
            v = f"\n\nHuman: {v}"
        if not v.endswith("\n\nAssistant:"):
            v = f"{v}\n\nAssistant:"
        return v


class CompletionResponse(BaseModel):
    completion: str
    stop_reason: str
    model: str



from typing import Any, Dict, Optional

from usepolvo.arms.base_graphql_client import BaseGraphQLClient
from usepolvo.arms.base_rate_limiter import BaseRateLimiter
from usepolvo.tentacles.linear.config import get_settings
from usepolvo.tentacles.linear.exceptions import handle_linear_error
from usepolvo.tentacles.linear.rate_limiter import LinearRateLimiter
from usepolvo.tentacles.linear.resources.issues.resource import LinearIssueResource


class LinearClient(BaseGraphQLClient):
    def __init__(self, api_key: Optional[str] = None):
        self.settings = get_settings()
        self.api_key = api_key if api_key else self.settings.linear_api_key
        self.base_url = self.settings.linear_base_url
        self.rate_limiter = LinearRateLimiter()
        self._issues = None
        super().__init__()

    def get_auth_headers(self) -> Dict[str, str]:
        """Provide Linear-specific authentication headers."""
        if self.api_key:
            return {"Authorization": f"{self.api_key}"}
        return {"Authorization": f"Bearer {self.api_key}"}

    @property
    def issues(self):
        if self._issues is None:
            self._issues = LinearIssueResource(self)
        return self._issues

    def handle_error(self, e):
        """Override error handling with Linear-specific logic."""
        raise handle_linear_error(e)

    def _get_resource_fields(self, resource_type: str) -> str:
        """Override to specify Linear-specific fields for each resource type."""
        if resource_type == "issue":
            return """
                title
                description
                state {
                    id
                    name
                }
                assignee {
                    id
                    name
                }
            """
        return super()._get_resource_fields(resource_type)

    @BaseRateLimiter.rate_limited
    def _request(self, method: str, endpoint: str, **kwargs) -> Dict[str, Any]:
        """
        Adapter method to maintain compatibility with REST-like interface.
        Translates REST-style requests into GraphQL queries.
        """
        query, variables = self._convert_rest_to_graphql(method, endpoint, **kwargs)
        return self.execute_query(query, variables)



from cachetools import TTLCache

from usepolvo.beak.config import get_settings


class BaseClient:
    def __init__(self):
        self.settings = get_settings()
        self.cache = TTLCache(maxsize=self.settings.cache_size, ttl=self.settings.cache_ttl)
        self.pagination_method = self.settings.pagination_method

    def handle_error(self, e):
        # Base error handling logic
        print(f"Error: {e}")



from usepolvo.arms.base_webhook import BaseWebhook


class StripeWebhook(BaseWebhook):
    def get_event_type(self, payload):
        return payload.get("type")

    async def default_handler(self, payload):
        """Default handler for unhandled Stripe webhook events."""
        print(f"Received unhandled Stripe webhook event: {payload.get('type')}")
        return {"status": "unhandled", "event_type": payload.get("type")}



from usepolvo.arms.base_rate_limiter import BaseRateLimiter


class ClaudeRateLimiter(BaseRateLimiter):
    def __init__(self, limit: int = 50):
        super().__init__()
        self.limit = limit
        self._initialize_window("requests")

    def wait_if_needed(self):
        with self.lock:
            self._wait_if_window_full("requests", self.limit, 60)  # 60 second window

    def get_limits(self):
        return {"requests_per_minute": self.limit}



from enum import Enum
from typing import Annotated, Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field, HttpUrl, model_validator


class WebhookAction(str, Enum):
    CREATE = "create"
    UPDATE = "update"
    REMOVE = "remove"
    RESTORE = "restore"


class WebhookType(str, Enum):
    ISSUE = "Issue"
    COMMENT = "Comment"
    PROJECT = "Project"
    CYCLE = "Cycle"
    REACTION = "Reaction"
    ISSUE_LABEL = "IssueLabel"
    ISSUE_ATTACHMENT = "IssueAttachment"
    PROJECT_UPDATE = "ProjectUpdate"
    DOCUMENT = "Document"
    USER = "User"


class UserType(str, Enum):
    USER = "user"
    BOT = "bot"
    UNKNOWN = "unknown"


class StateType(str, Enum):
    TRIAGE = "triage"
    BACKLOG = "backlog"
    UNSTARTED = "unstarted"
    STARTED = "started"
    COMPLETED = "completed"
    CANCELED = "canceled"
    DUPLICATE = "duplicate"


# Base Models
class User(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    email: Optional[str] = None
    avatarUrl: Optional[HttpUrl] = None
    type: Optional[UserType] = None
    active: Optional[bool] = None
    displayName: Optional[str] = None
    organizationId: Optional[str] = None


class Label(BaseModel):
    id: Optional[str] = None
    color: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None


class Project(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    url: Optional[HttpUrl] = None
    description: Optional[str] = None
    createdAt: Optional[str] = None
    updatedAt: Optional[str] = None
    state: Optional[str] = None
    startDate: Optional[str] = None
    targetDate: Optional[str] = None


class State(BaseModel):
    id: Optional[str] = None
    color: Optional[str] = None
    name: Optional[str] = None
    type: Optional[StateType] = None
    position: Optional[float] = None


class Team(BaseModel):
    id: Optional[str] = None
    key: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None


class Attachment(BaseModel):
    id: Optional[str] = None
    title: Optional[str] = None
    url: Optional[HttpUrl] = None
    size: Optional[int] = None
    contentType: Optional[str] = None
    uploadUrl: Optional[HttpUrl] = None


# Common base class for all webhook data types
class BaseWebhookData(BaseModel):
    """Base class containing fields common to all webhook types"""

    id: str
    createdAt: str
    updatedAt: str
    type: WebhookType  # Discriminator field


# Event-specific Data Models
class CommonOptionalFields(BaseModel):
    """Common optional fields that appear in multiple event types"""

    number: Optional[int] = None
    title: Optional[str] = None
    priority: Optional[int] = None
    boardOrder: Optional[float] = None
    sortOrder: Optional[float] = None
    prioritySortOrder: Optional[float] = None
    startedAt: Optional[str] = None
    completedAt: Optional[str] = None
    slaType: Optional[str] = None
    addedToProjectAt: Optional[str] = None
    labelIds: Optional[List[str]] = None
    teamId: Optional[str] = None
    projectId: Optional[str] = None
    lastAppliedTemplateId: Optional[str] = None
    previousIdentifiers: Optional[List[str]] = None
    creatorId: Optional[str] = None
    assigneeId: Optional[str] = None
    stateId: Optional[str] = None
    reactionData: Optional[List[Dict[str, Any]]] = None
    priorityLabel: Optional[str] = None
    botActor: Optional[User] = None
    identifier: Optional[str] = None
    url: Optional[HttpUrl] = None
    subscriberIds: Optional[List[str]] = None
    assignee: Optional[User] = None
    project: Optional[Project] = None
    state: Optional[State] = None
    team: Optional[Team] = None
    labels: Optional[List[Label]] = None
    description: Optional[str] = None
    descriptionData: Optional[str] = None


class IssueData(BaseWebhookData, CommonOptionalFields):
    """Issue-specific webhook data"""

    type: Literal[WebhookType.ISSUE]


class CommentData(BaseWebhookData, CommonOptionalFields):
    """Comment-specific webhook data"""

    type: Literal[WebhookType.COMMENT]
    body: Optional[str] = None
    userId: Optional[str] = None
    issueId: Optional[str] = None
    bodyData: Optional[str] = None


class IssueLabelData(BaseWebhookData, CommonOptionalFields):
    """IssueLabel-specific webhook data"""

    type: Literal[WebhookType.ISSUE_LABEL]


class IssueAttachmentData(BaseWebhookData):
    """IssueAttachment-specific webhook data"""

    type: Literal[WebhookType.ISSUE_ATTACHMENT]
    issueId: Optional[str] = None
    attachment: Optional[Attachment] = None


class ProjectData(BaseWebhookData):
    """Project-specific webhook data"""

    type: Literal[WebhookType.PROJECT]
    name: Optional[str] = None
    description: Optional[str] = None
    teamId: Optional[str] = None
    state: Optional[str] = None
    startDate: Optional[str] = None
    targetDate: Optional[str] = None
    lead: Optional[User] = None


class ProjectUpdateData(BaseWebhookData):
    """ProjectUpdate-specific webhook data"""

    type: Literal[WebhookType.PROJECT_UPDATE]
    projectId: Optional[str] = None
    userId: Optional[str] = None
    body: Optional[str] = None
    bodyData: Optional[str] = None


class CycleData(BaseWebhookData):
    """Cycle-specific webhook data"""

    type: Literal[WebhookType.CYCLE]
    name: Optional[str] = None
    number: Optional[int] = None
    teamId: Optional[str] = None
    startDate: Optional[str] = None
    endDate: Optional[str] = None


class ReactionData(BaseWebhookData):
    """Reaction-specific webhook data"""

    type: Literal[WebhookType.REACTION]
    emoji: Optional[str] = None
    userId: Optional[str] = None
    commentId: Optional[str] = None
    issueId: Optional[str] = None


class DocumentData(BaseWebhookData):
    """Document-specific webhook data"""

    type: Literal[WebhookType.DOCUMENT]
    title: Optional[str] = None
    content: Optional[str] = None
    contentData: Optional[str] = None
    icon: Optional[str] = None
    organizationId: Optional[str] = None


class UserData(BaseWebhookData):
    """User-specific webhook data"""

    type: Literal[WebhookType.USER]
    name: Optional[str] = None
    email: Optional[str] = None
    active: Optional[bool] = None
    displayName: Optional[str] = None
    organizationId: Optional[str] = None
    avatarUrl: Optional[HttpUrl] = None


# Discriminated union of all possible webhook data types
WebhookData = Annotated[
    Union[
        IssueData,
        CommentData,
        IssueLabelData,
        IssueAttachmentData,
        ProjectData,
        ProjectUpdateData,
        CycleData,
        ReactionData,
        DocumentData,
        UserData,
    ],
    Field(discriminator="type"),
]


class UpdatedFromData(BaseModel):
    """Represents the partial data in updatedFrom field"""

    updatedAt: Optional[str] = None
    sortOrder: Optional[float] = None
    prioritySortOrder: Optional[float] = None
    completedAt: Optional[str] = None
    stateId: Optional[str] = None
    assigneeId: Optional[str] = None
    priority: Optional[int] = None
    title: Optional[str] = None
    description: Optional[str] = None
    teamId: Optional[str] = None


class LinearWebhookPayload(BaseModel):
    """Main webhook payload model"""

    action: WebhookAction
    type: WebhookType
    data: WebhookData
    actor: Optional[User] = None
    createdAt: Optional[str] = None
    updatedFrom: Optional[UpdatedFromData] = None
    url: Optional[HttpUrl] = None
    organizationId: str
    webhookTimestamp: int
    webhookId: str

    @model_validator(mode="before")
    @classmethod
    def inherit_webhook_type(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Ensures the type field is properly inherited by the data object"""
        if isinstance(values, dict):
            webhook_type = values.get("type")
            if webhook_type and "data" in values and isinstance(values["data"], dict):
                # Inherit the type from the parent webhook
                values["data"]["type"] = webhook_type
        return values

    def get_event_type(self) -> str:
        """Returns event type in format 'type.action' (e.g., 'issue.update')"""
        return f"{self.type.lower()}.{self.action}"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LinearWebhookPayload":
        """Helper method to create instance from dictionary"""
        # Ensure the type field is added to the data for discrimination
        if "data" in data and "type" in data:
            data["data"]["type"] = data["type"]
        return cls(**data)

    def get_identifier(self) -> str:
        """Returns the unique identifier for the webhook data"""
        if isinstance(self.data, IssueData) and self.data.identifier:
            return self.data.identifier
        return self.data.id


# Example usage:
"""
# Parse a webhook payload
payload = LinearWebhookPayload.from_dict({
    "action": "create",
    "type": "IssueAttachment",
    "data": {
        "id": "123",
        "createdAt": "2025-01-17T16:57:04.664Z",
        "updatedAt": "2025-01-17T16:57:04.664Z",
        "issueId": "issue-123",
        "attachment": {
            "title": "Screenshot.png",
            "url": "https://example.com/files/screenshot.png"
        }
    },
    "organizationId": "org-123",
    "webhookTimestamp": 1737133024760,
    "webhookId": "webhook-123"
})

# Type-safe access to data
if isinstance(payload.data, IssueAttachmentData):
    print(f"Attachment title: {payload.data.attachment.title}")
elif isinstance(payload.data, DocumentData):
    print(f"Document title: {payload.data.title}")
"""



from typing import Any, Dict

from google.api_core import exceptions as google_exceptions

from usepolvo.arms.base_resource import BaseResource
from usepolvo.beak.exceptions import ResourceNotFoundError, ValidationError
from usepolvo.tentacles.gemini.exceptions import GeminiError


class GeminiCompletionResource(BaseResource):
    def __init__(self, client):
        super().__init__(client)
        self.genai = client.client

    def create(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Create a new completion using Gemini.

        :param data: The data for the new completion
        :return: The created completion object
        :raises ValidationError: If the data is invalid
        """
        try:
            self.client.rate_limiter.wait_if_needed()
            model = self.genai.GenerativeModel(data.pop("model"))
            response = model.generate_content(data.pop("prompt"), **data)
            return response.text
        except google_exceptions.InvalidArgument as e:
            raise ValidationError(f"Invalid data for creating completion: {str(e)}")
        except Exception as e:
            self.client.handle_error(e)

    def get(self, completion_id: str) -> Dict[str, Any]:
        pass

    def list(self, **kwargs) -> Dict[str, Any]:
        pass

    def update(self, completion_id: str, data: Dict[str, Any]) -> Dict[str, Any]:
        pass

    def delete(self, completion_id: str) -> Dict[str, Any]:
        pass



from typing import Any, Dict

from pydantic import BaseModel


class CertnWebhookPayload(BaseModel):
    created: str
    modified: str
    last_updated: str
    submitted_time: str
    request_identity_verification: bool = False
    request_equifax: bool = False
    request_base: bool = False
    request_instant_verify_employment: bool = False
    request_instant_verify_education: bool = False
    request_instant_verify_credential: bool = False
    request_international_criminal_record_check: bool = False
    request_enhanced_identity_verification: bool = False
    request_motor_vehicle_records: bool = False
    request_criminal_record_check: bool = False
    request_enhanced_criminal_record_check: bool = False
    request_vulnerable_sector_criminal_record_check: bool = False
    request_employer_references: bool = False
    request_address_references: bool = False
    request_employer_phone_references: bool = False
    request_address_phone_references: bool = False
    request_softcheck: bool = False
    request_social_media_check: bool = False
    request_soquij: bool = False
    request_us_criminal_record_check_tier_1: bool = False
    request_us_criminal_record_check_tier_2: bool = False
    request_us_criminal_record_check_tier_3: bool = False
    request_education_verification: bool = False
    request_credential_verification: bool = False
    request_employment_verification: bool = False
    request_vaccination_check: bool = False
    request_australian_criminal_intelligence_commission_check: bool = False
    request_right_to_work: bool = False
    request_uk_basic_dbs_check: bool = False
    request_uk_basic_ds_check: bool = False
    request_uk_right_to_work_check: bool = False

    def get_event_type(self) -> str:
        for field, value in self.model_dump().items():
            if field.startswith("request_") and value is True:
                return field
        return "unknown"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "CertnWebhookPayload":
        return cls(**data)
